from keras.models import Model
from keras.layers import TimeDistributed, Embedding, Input, LSTM, Bidirectional, concatenate, Dense, average
from keras.initializers import RandomUniform
from ELMo import ElmoEmbeddingLayer
from keras_contrib.layers import CRF


def NeuroTPR_initiate(wordEmbeddings, posEmbeddings, params):
    words_input = Input(shape=(None,), dtype='int32', name='words_input')
    words = Embedding(input_dim=wordEmbeddings.shape[0], output_dim=wordEmbeddings.shape[1], weights=[wordEmbeddings],
                      trainable=False)(words_input)

    pos_input = Input(shape=(None,), dtype='int32', name='pos_input')
    pos = Embedding(input_dim=posEmbeddings.shape[0], output_dim=posEmbeddings.shape[1], weights=[posEmbeddings],
                    trainable=False)(pos_input)

    character_input1 = Input(shape=(None, 52), name='char_input_normal')
    embed_char_out1 = TimeDistributed(Embedding(input_dim=params["char_length"], output_dim=50, embeddings_initializer=
                                     RandomUniform(minval=-0.5, maxval=0.5)), name='char_embedding1')(character_input1)
    char_lstm1 = TimeDistributed(Bidirectional(LSTM(25, return_sequences=False, return_state=False, recurrent_dropout=0.25)),
                                name='char_LSTM1')(embed_char_out1)

    character_input2 = Input(shape=(None, 52), name='char_input_caseless')
    embed_char_out2 = TimeDistributed(Embedding(input_dim=params["char_length_caseless"], output_dim=50, embeddings_initializer=
                                     RandomUniform(minval=-0.5, maxval=0.5)), name='char_embedding2')(character_input2)
    char_lstm2 = TimeDistributed(Bidirectional(LSTM(25, return_sequences=False, return_state=False, recurrent_dropout=0.25)),
                                name='char_LSTM2')(embed_char_out2)

    char_lstm3 = average([char_lstm1, char_lstm2])

    words_elmo_input = Input(shape=(1,), dtype='string', name='words_elmo_input')
    words_elmo = ElmoEmbeddingLayer(trainable=False)(words_elmo_input)

    output = concatenate([words, pos, char_lstm3, words_elmo], axis=-1)

    word_lstm = Bidirectional(LSTM(60, return_sequences=True, dropout=0.50, recurrent_dropout=0.25), name='word_LSTM')(output)

    fc1 = TimeDistributed(Dense(50),name='fully_connected1')(word_lstm)

    my_crf = CRF(params["label_length"], sparse_target=True, name='CRF_layer')(fc1)

    model = Model(inputs=[words_input, character_input1, character_input2, pos_input, words_elmo_input], outputs=[my_crf]) # 

    return model


from keras import backend as K
from keras.engine.topology import Layer
import tensorflow as tf
import tensorflow_hub as hub


class ElmoEmbeddingLayer(Layer):
    def __init__(self, trainable, **kwargs):
        self.dimensions = 1024
        self.trainable = trainable
        super(ElmoEmbeddingLayer, self).__init__(**kwargs)

    def build(self, input_shape):
            self.elmo = hub.Module('https://tfhub.dev/google/elmo/2', trainable=self.trainable, name="{}_module".format(self.name))
            self.trainable_weights += tf.compat.v1.trainable_variables(scope="^{}_module/.*".format(self.name))
            super(ElmoEmbeddingLayer, self).build(input_shape)

    def call(self, x, mask=None):
            result = self.elmo(K.squeeze(K.cast(x, tf.string), axis=1),
                          as_dict=True,
                          signature='default',
                          )['elmo']
            return result

    # def compute_mask(self, inputs, mask=None):
    #         return K.not_equal(inputs, '--PAD--')

    def compute_output_shape(self, input_shape):
            return (input_shape[0], None, self.dimensions)

    def get_config(self):
        config = {'trainable': self.trainable}
        base_config = super(ElmoEmbeddingLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))





import nltk
from nltk.tag import StanfordNERTagger
from nltk import RegexpParser
from nltk import conlltags2tree, tree2conlltags

# nltk.download('averaged_perceptron_tagger')


def sentence_to_file(filename="../trainingdata/temp.txt", psentence=[]):
    with open(filename, "a") as wf:
        for item in psentence:
            wf.write(item[0]+"\t"+item[1]+"\n")

        wf.write("\n")
        wf.close()


def tweet_ner_tagger(text_list, st, cp):
    text = ["URL" if word[0].startswith("http") else word[0] for word in text_list]
    gold_tag = [word[1] for word in text_list]

    tokenized_text = text

    ner_taggers = st.tag(tokenized_text)
    pos_taggers = nltk.pos_tag(tokenized_text)
    chunk_taggers = tree2conlltags(cp.parse(pos_taggers))

    ner_sequence = [item[1] for item in ner_taggers]
    pos_sequence = [item[1] for item in pos_taggers]
    chunking_sequence = [item[1] for item in chunk_taggers]

    return text, gold_tag, ner_sequence, pos_sequence, chunking_sequence


def iterate_file(filename_read, filename_write, st, cp):
    '''
    read file
    return format :
    [ ['EU', 'B-ORG'], ['rejects', 'O'], ['German', 'B-MISC'], ['call', 'O'], ['to', 'O'], ['boycott', 'O'], ['British', 'B-MISC'], ['lamb', 'O'], ['.', 'O'] ]
    '''
    read_f = open(filename_read, "r+")
    write_f = open(filename_write, "w+")
    j = 0
    sentence = []
    for line in read_f:
        line = line.rstrip()
        if len(line) == 0:
            # if len(sentence) > 0 and len(sentence) < 25:
            # Here to add the linguistics features
            c1, c2, c3, c4, c5 = tweet_ner_tagger(sentence, st, cp)
            if len(c1) != len(c3):
                sentence_to_file(psentence=sentence)
                sentence = []
                print("drop!!!!")
                continue
            for i in range(0, len(c1)):
                write_f.write(c1[i] + "\t" + c2[i] + "\t" + c3[i] + "\t" + c4[i] + "\t" + c5[i] + "\n")
            write_f.write("\n")

            sentence = []
            j += 1
            print(j)
            continue

        splits = line.split('\t')
        sentence.append([splits[0], splits[-1]])

    if len(sentence) > 0:
        c1, c2, c3, c4, c5 = tweet_ner_tagger(sentence, st)
        for i in range(0, len(c1)):
            write_f.write(c1[i] + "\t" + c2[i] + "\t" + c3[i] + "\t" + c4[i] + "\t" + c5[i] + "\n")
        write_f.write("\n")
        sentence = []

    write_f.close()
    read_f.close()
    print("OVER!!!")


if __name__ == "__main__":
    raw_dataset_path = "../trainingdata/wnut17train2.txt"
    processed_dataset_path = "../trainingdata/wnut17train2_add_features.txt"



    stanford_tagger = StanfordNERTagger('../stanford-ner/classifiers/english.all.3class.distsim.crf.ser.gz',
                                        '../stanford-ner/stanford-ner.jar',
                                        encoding='utf-8')

    grammar = "NP: {<DT>?<JJ>*<NN>}"
    cp = RegexpParser(grammar)

    iterate_file(raw_dataset_path, processed_dataset_path, stanford_tagger, cp)







import numpy as np
from preprocess import readfile, createBatches, createMatrices_char, iterate_minibatches_char, addCharInformatioin,\
    padding
import keras.backend as K
from keras.utils import Progbar
from keras_contrib.losses import crf_loss
from keras_contrib.metrics import crf_viterbi_accuracy
from NeuroTPR import NeuroTPR_initiate
from DM_NLP import DMNLP_initiate


epochs1 = 35
epochs2 = 20
embedding_choice = 0 # 0: Twitter_Glove; 1: Normal_Glove
Model_DIR = 'Change with your own Root Directory'


def get_ner_embedding():
    ner2Idx = {'LOCATION': 0, 'ORGANIZATION': 1, 'PERSON': 2, 'O': 3, 'PADDING_TOKEN': 4}
    return ner2Idx

def get_char_embedding():
    char2Idx = {"PADDING": 0, "UNKNOWN": 1}
    char2Idx_caseless = {"PADDING": 0, "UNKNOWN": 1}

    for c in " 0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.,-_()[]{}!?:;#'\"/\\%$`&=*+@^~|":
        char2Idx[c] = len(char2Idx)

    for c in " 0123456789abcdefghijklmnopqrstuvwxyz.,-_()[]{}!?:;#'\"/\\%$`&=*+@^~|":
        char2Idx_caseless[c] = len(char2Idx_caseless)

    for c in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        char2Idx_caseless[c] = char2Idx_caseless[c.lower()]

    return char2Idx, char2Idx_caseless


def save_keras_model(model, modelDIR, name1="/outputs/NeuroTPR.json", name2="/outputs/NeuroTPR.h5"):
    model_json = model.to_json()
    with open(modelDIR + name1, "w") as json_file:
        json_file.write(model_json)
    model.save_weights(modelDIR + name2)
    print("Saved model to disk")


if __name__ == '__main__':
	# Read training file as txt
    trainSentences1 = readfile("../data/WikiTPR3000_train_add_features.txt")
    trainSentences2 = readfile("../data/wnut17train2_add_features.txt")
    testSentences = readfile("../data/Harvey1000.txt")
    print("Finishing reading the training data from file! ")

    # Process the character features from words
    trainSentences1 = addCharInformatioin(trainSentences1)
    trainSentences2 = addCharInformatioin(trainSentences2)
    print("Finishing adding the character information!!!")

    labelSet = set()
    posSet = set()
    all_words = {}

    for TrainSentences in [trainSentences1, trainSentences2]:
    	for sentence in TrainSentences:
	        for token, char, label, ner, pos in sentence:
	            labelSet.add(label)
	            posSet.add(pos)
	            all_words[token.lower()] = True

    for sentence in testSentences:
        for token, label, ner, pos in sentence:
            all_words[token.lower()] = True

    print("Finishing cleaning the raw text! ")

    # -------------- Process the linguistic embedding ------------ #
    # Create character feature lookup dictionary
    char2Idx, char2Idx_caseless = get_char_embedding()

    # Create a mapping for the labels
    label2Idx = {}
    for label in labelSet:
        label2Idx[label] = len(label2Idx)

    idx2Label = {v: k for k, v in label2Idx.items()}
    print(idx2Label)

    # Create pos feature lookup dictionary
    pos2Idx = {}
    for pos in posSet:
        pos2Idx[pos] = len(pos2Idx)

    pos2Idx["UNKNOWN"] = len(pos2Idx)
    posEmbeddings = np.identity(len(pos2Idx), dtype='float32')

    # Create ner feature lookup dictionary
    ner2Idx = get_ner_embedding()
    nerEmbeddings = np.identity(len(ner2Idx), dtype='float32')

    # -------------- Process the word embedding ------------ #
    print("Begin to add the word embedding information! ")
    word2Idx = {}
    wordEmbeddings = []

    if embedding_choice == 0:
        fEmbeddings = open("../embedding/glove.twitter.27B.200d.txt", encoding="utf-8")
    else:
        fEmbeddings = open("../embedding/glove.6B.300d.txt", encoding="utf-8")

    for line in fEmbeddings:
        split = line.strip().split(" ")

        if len(word2Idx) == 0:  # Add padding+unknown
            word2Idx["PADDING_TOKEN"] = len(word2Idx)
            vector = np.zeros(len(split) - 1)  # Zero vector vor 'PADDING' word
            wordEmbeddings.append(vector)

            word2Idx["UNKNOWN_TOKEN"] = len(word2Idx)
            vector = np.random.uniform(-0.25, 0.25, len(split) - 1)
            wordEmbeddings.append(vector)

        if split[0].lower() in all_words:
            vector = np.array([float(num) for num in split[1:]])
            wordEmbeddings.append(vector)
            word2Idx[split[0]] = len(word2Idx)

    wordEmbeddings = np.array(wordEmbeddings)
    print("Finishing constructing word-vectors dictionary! ")

    # Save all embeddings and linguistic features lookup into files
    np.save("outputs/idx2Label.npy", idx2Label)
    np.save("outputs/word2Idx.npy", word2Idx)
    np.save("outputs/char2Idx.npy", char2Idx)
    np.save("outputs/ner2Idx.npy", ner2Idx)
    np.save("outputs/char2Idx_caseless.npy", char2Idx_caseless)
    np.save("outputs/pos2Idx.npy", pos2Idx)

    # ----------------------------------------#

    # Convert training senetences into Keras-compatitable format
    train_set1 = padding(createMatrices_char(trainSentences1, word2Idx, label2Idx, char2Idx, char2Idx_caseless, pos2Idx))
    train_set2 = padding(createMatrices_char(trainSentences2, word2Idx, label2Idx, char2Idx, char2Idx_caseless, pos2Idx))

    # ------------- Build up NeuroTPR model architecture -----------------#

    p = {"label_length": len(idx2Label), "char_length": len(char2Idx), "char_length_caseless": len(char2Idx)-26}
    model = NeuroTPR_initiate(wordEmbeddings, posEmbeddings, params=p)

    model.compile(optimizer='nadam', loss=crf_loss, metrics=[crf_viterbi_accuracy])
    model.summary()

    # -----------------  Model Built Up Finished -------------------- #

    # ----------------- Model Training and Weights Saving ----------------- #

    # Training senetences to batches

    train_batch1, train_batch_len1 = createBatches(train_set1)
    train_batch2, train_batch_len2 = createBatches(train_set2)

    for epoch in range(epochs1):
        train_accuracy = 0.0
        print("Epoch %d/%d" % (epoch, epochs))
        a = Progbar(len(train_batch_len1))
        for i, batch in enumerate(iterate_minibatches_char(train_batch1, train_batch_len1)):
            labels, tokens, chars, chars2, poss, strings = batch
            results = model.train_on_batch([tokens, chars, chars2, poss, strings], labels)
            train_accuracy += results[1]
            a.update(i)
        a.update(i + 1)
        print(train_accuracy/(i+1))
        if train_accuracy/(i+1) > 0.99:
            break

    for epoch in range(epochs2):
        train_accuracy = 0.0
        print("Epoch %d/%d" % (epoch, epochs))
        a = Progbar(len(train_batch_len2))
        for i, batch in enumerate(iterate_minibatches_char(train_batch2, train_batch_len2)):
            labels, tokens, chars, chars2, poss, strings = batch
            results = model.train_on_batch([tokens, chars, chars2, poss, strings], labels) # 
            train_accuracy += results[1]
            a.update(i)
        a.update(i + 1)
        print(train_accuracy/(i+1))

    # -------------- Save moddel -----------------#

    model_name_j = "/outputs/NeuroTPR.json"
    model_name_h = "/outputs/NeuroTPR.h5"

    save_keras_model(model, Model_DIR, name1=model_name_j, name2=model_name_h)



import numpy as np
import copy
import json
from nltk import pos_tag
from nltk.tokenize import TweetTokenizer
from emoji import get_emoji_regexp
from neurotpr.ELMo import ElmoEmbeddingLayer
from keras.preprocessing.sequence import pad_sequences
from keras.models import model_from_json
from keras_contrib.layers import CRF
from keras_contrib.losses import crf_loss
from keras_contrib.metrics import crf_viterbi_accuracy
import nltk
nltk.download('averaged_perceptron_tagger') # download the ner tagger used by nltk




#print("Loading the word and character embeddings...")
word2Idx = None
labelIdx = None
char2Idx = None
char2Idx_caseless = None
pos2Idx = None
inv_labelIdx = None
bloc_index = None
iloc_index = None

tpr_model = None



def load_model(model_path):
    
    global tpr_model
    
    json_file = open(model_path + 'NeuroTPR.json', 'r')
    loaded_model_json = json_file.read()
    json_file.close()
    tpr_model = model_from_json(loaded_model_json, custom_objects={
        'CRF': CRF, 'crf_loss': crf_loss, "crf_viterbi_accuracy": crf_viterbi_accuracy,
        'ElmoEmbeddingLayer': ElmoEmbeddingLayer})
    tpr_model.load_weights(model_path + 'NeuroTPR.h5')
    

    
    global word2Idx 
    global labelIdx 
    global char2Idx 
    global char2Idx_caseless 
    global pos2Idx 
    global inv_labelIdx 
    global bloc_index 
    global iloc_index 
    
    wordembedding_path = model_path+"word2Idx.npy"
    labelset_path = model_path+"idx2Label.npy"
    charembedding_path = model_path+"char2Idx.npy"
    charembedding_path2 = model_path+"char2Idx_caseless.npy"
    posembedding_path = model_path+"pos2Idx.npy"

    word2Idx = np.load(wordembedding_path, allow_pickle=True).item()
    labelIdx = np.load(labelset_path, allow_pickle=True).item()
    char2Idx = np.load(charembedding_path, allow_pickle=True).item()
    char2Idx_caseless = np.load(charembedding_path2, allow_pickle=True).item()
    pos2Idx = np.load(posembedding_path, allow_pickle=True).item()
    inv_labelIdx = {v: k for k, v in labelIdx.items()}
    bloc_index = inv_labelIdx["B-location"]
    iloc_index = inv_labelIdx["I-location"]
    
    
    print("Loaded pretrained NeuroTPR model")
    



def delete_emoji(raw_text):
    return get_emoji_regexp().sub(r'', raw_text)


def tweet_pos_tagger(text_list):
    word_idx = []
    start_idx = 1
    for word in text_list:
        word_idx.append(start_idx)
        start_idx += len(word)+1

    tokenized_text = ["URL" if word.startswith("http") else word for word in text_list]
    pos_taggers = pos_tag(tokenized_text)
    pos_sequence = [item[1] for item in pos_taggers]

    return tokenized_text, word_idx, pos_sequence


def preprocess_tweet(text):
    text = delete_emoji(text)
    tweet_token = TweetTokenizer(preserve_case=True, strip_handles=True, reduce_len=True)
    results = tweet_token.tokenize(text)
    results = [item[1:] if (item.startswith("#") and len(item)>1) else item for item in results]

    tokens, indexes, poss = tweet_pos_tagger(results)
    new_tweets = []
    for i in range(0, len(tokens)):
        new_tweets.append([tokens[i], indexes[i], poss[i]])

    return new_tweets


def addCharInformatioin(Sentence):
    for i, data in enumerate(Sentence):
            chars = [c for c in data[0]]
            Sentence[i] = [data[0], chars, data[1], data[2]]
    return Sentence


def createMatrices_nolabel_char(sentence, word2Idx, char2Idx, char2Idx_caseless, pos2Idx):
    unknownIdx = word2Idx['UNKNOWN_TOKEN']
    paddingIdx = word2Idx['PADDING_TOKEN']

    all_features = []

    wordCount = 0
    unknownWordCount = 0

    wordIndices = []
    charIndices1 = []
    charIndices2 = []
    posIndices = []
    wordStrings = ""

    for word, char, index, pos in sentence:
        wordCount += 1
        wordStrings = wordStrings + word + " "

        if word in word2Idx:
            wordIdx = word2Idx[word]
        elif word.lower() in word2Idx:
            wordIdx = word2Idx[word.lower()]
        else:
            wordIdx = unknownIdx
            unknownWordCount += 1
        charIdx1 = []
        charIdx2 = []
        for x in char:
            if x in char2Idx:
                charIdx1.append(char2Idx[x])
                charIdx2.append(char2Idx_caseless[x])
            else:
                charIdx1.append(char2Idx['UNKNOWN'])
                charIdx2.append(char2Idx_caseless['UNKNOWN'])

        # Get the label and map to int
        wordIndices.append(wordIdx)
        charIndices1.append(charIdx1)
        charIndices2.append(charIdx2)

        if pos in pos2Idx:
            posIndices.append(pos2Idx[pos])
        else:
            posIndices.append(pos2Idx['UNKNOWN'])

    all_features.append([wordIndices, charIndices1, charIndices2, posIndices, wordStrings[:-1]])

    return all_features


def padding(Sentences):
    maxlen = 52
    for sentence in Sentences:
        char = sentence[1]
        for x in char:
            maxlen = max(maxlen, len(x))
    for i, sentence in enumerate(Sentences):
        Sentences[i][1] = pad_sequences(Sentences[i][1], 52, padding='post')
        Sentences[i][2] = pad_sequences(Sentences[i][2], 52, padding='post')
    return Sentences


def build_senteceMatrix(sentence):

    features = []
    wordIndices = []
    indexIndices = []
    for word, label, pos in sentence:
        # Get the label and map to int
        wordIndices.append(word)
        indexIndices.append(int(label))
    features.append([wordIndices, indexIndices])

    return features


def geoparse_tweet(tweet_features, tweet_raw, model, iinx, binx):
    toponyms_set = []

    for i, sentence in enumerate(tweet_features):
        tokens, chars, chars2, poss, text = sentence
        tokens = np.asarray([tokens])
        chars = np.asarray([chars])
        chars2 = np.asarray([chars2])
        poss = np.asarray([poss])
        text = np.asarray([text])

        if len(text[0]) < 5:
            print("Too short input for the model")
            break

        pred = model.predict([tokens, chars, chars2, poss, text], verbose=False)[0]  #

        pred = pred.argmax(axis=-1)  # Predict the classes

        tokenIdx = 0
        toponym_item = {}
        while tokenIdx < len(pred):
            toponym = ""
            if pred[tokenIdx] == binx:  # A new chunk starts
                old_idx = tokenIdx
                toponym = tweet_raw[i][0][tokenIdx]
                tokenIdx += 1

                while tokenIdx < len(pred) and pred[tokenIdx] == iinx:
                    toponym += (" " + tweet_raw[i][0][tokenIdx])
                    tokenIdx += 1

                if toponym.endswith(","):
                    toponym = toponym[:-2]

                if toponym.find("Harvey") == -1 and len(toponym) > 1:
                    toponym_item["location_name"] = toponym
                    toponym_item["start_idx"] = tweet_raw[i][1][old_idx]
                    toponym_item["end_idx"] = tweet_raw[i][1][old_idx] + len(toponym) - 1
                    toponyms_set.append(dict(toponym_item))

            elif pred[tokenIdx] == iinx:
                tokenIdx += 1

            else:
                tokenIdx += 1

    return toponyms_set



   

def topo_recog(tweet):
    
    
    processed_tweet = preprocess_tweet(tweet)

    processed_tweet_bak = copy.deepcopy(processed_tweet)

    processed_tweet_char = addCharInformatioin(processed_tweet)

    tweet_all_features = padding(createMatrices_nolabel_char(processed_tweet_char, word2Idx, char2Idx, char2Idx_caseless, pos2Idx))
    tweet_word_index = build_senteceMatrix(processed_tweet_bak)

    result_json = geoparse_tweet(tweet_all_features, tweet_word_index, tpr_model, iloc_index, bloc_index)
    result_json = json.dumps(result_json)

    return(result_json)
    



    
    #print("Initialization finishes...")





from keras.preprocessing.sequence import pad_sequences

def build_senteceMatrix(sentences):

    dataset = []
    for sentence in sentences:
        wordIndices = []
        indexIndices = []
        for word, label, ner, pos in sentence:
            # Get the label and map to int
            wordIndices.append(word)
            indexIndices.append(int(label))
        dataset.append([wordIndices, indexIndices])

    return dataset


def createMatrices_nolabel(sentences, word2Idx):
    unknownIdx = word2Idx['UNKNOWN_TOKEN']
    paddingIdx = word2Idx['PADDING_TOKEN']

    dataset = []

    for sentence in sentences:
        wordIndices = []
        indexIndices = []

        for word, label in sentence:
            if word in word2Idx:
                wordIdx = word2Idx[word]
            elif word.lower() in word2Idx:
                wordIdx = word2Idx[word.lower()]
            else:
                wordIdx = unknownIdx

            # Get the label and map to int
            wordIndices.append(wordIdx)
            indexIndices.append(str(label))
        dataset.append([wordIndices, indexIndices])

    return dataset


def readfile_nolabel(filename):
    '''
    read file
    return format :
    [ ['EU', 'B-ORG'], ['rejects', 'O'], ['German', 'B-MISC'], ['call', 'O'], ['to', 'O'], ['boycott', 'O'], ['British', 'B-MISC'], ['lamb', 'O'], ['.', 'O'] ]
    '''
    f = open(filename)
    sentences = []
    sentence = []
    current_index = 1
    for line in f:
        line = line.rstrip()
        if len(line) == 0:
            if len(sentence) > 0:
                sentences.append(sentence)
                sentence = []
            continue
        splits = line.split('\t')
        if len(splits[0]) == 0:
            continue
        sentence.append([splits[0], current_index, splits[1], splits[2]])
        current_index += (1 + len(splits[0]))

    if len(sentence) > 0:
        sentences.append(sentence)
        sentence = []

    return sentences


def createMatrices_nolabel_char(sentences, word2Idx, char2Idx, char2Idx_caseless, pos2Idx):
    unknownIdx = word2Idx['UNKNOWN_TOKEN']
    paddingIdx = word2Idx['PADDING_TOKEN']

    dataset = []

    wordCount = 0
    unknownWordCount = 0

    for sentence in sentences:
        wordIndices = []
        charIndices1 = []
        charIndices2 = []
        posIndices = []
        wordStrings = ""

        for word, char, label, pos in sentence:
            wordCount += 1
            wordStrings = wordStrings + word + " "

            if word in word2Idx:
                wordIdx = word2Idx[word]
            elif word.lower() in word2Idx:
                wordIdx = word2Idx[word.lower()]
            else:
                wordIdx = unknownIdx
                unknownWordCount += 1
            charIdx1 = []
            charIdx2 = []
            for x in char:
                if x in char2Idx:
                    charIdx1.append(char2Idx[x])
                    charIdx2.append(char2Idx_caseless[x])
                else:
                    charIdx1.append(char2Idx['UNKNOWN'])
                    charIdx2.append(char2Idx_caseless['UNKNOWN'])

            # Get the label and map to int
            wordIndices.append(wordIdx)
            charIndices1.append(charIdx1)
            charIndices2.append(charIdx2)
            if pos in pos2Idx:
                posIndices.append(pos2Idx[pos])
            else:
                posIndices.append(pos2Idx['UNKNOWN'])

        dataset.append([wordIndices, charIndices1, charIndices2, posIndices, wordStrings[:-1]])

    return dataset


def addCharInformatioin(Sentences):
    for i, sentence in enumerate(Sentences):
        for j, data in enumerate(sentence):
            chars = [c for c in data[0]]
            Sentences[i][j] = [data[0], chars, data[1], data[2], data[3]]
    return Sentences


def padding(Sentences):
    maxlen = 52
    for sentence in Sentences:
        char = sentence[1]
        for x in char:
            maxlen = max(maxlen, len(x))
    for i, sentence in enumerate(Sentences):
        Sentences[i][1] = pad_sequences(Sentences[i][1], 52, padding='post')
        Sentences[i][2] = pad_sequences(Sentences[i][2], 52, padding='post')
    return Sentences



import numpy as np
import copy
import os
from re import finditer
from neurotpr.ELMo import ElmoEmbeddingLayer
from keras.models import model_from_json
from keras_contrib.layers import CRF
from keras_contrib.losses import crf_loss
from keras_contrib.metrics import crf_viterbi_accuracy
from neurotpr.prediction_pre import readfile_nolabel, createMatrices_nolabel_char, build_senteceMatrix, padding, addCharInformatioin

model_path = "../Model/outputs/"
file_path = "../data/testing/HarveyTweet.txt"
wordembedding_path = "../Model/outputs/word2Idx.npy"
labelset_path = "../Model/outputs/idx2Label.npy"
charembedding_path = "../Model/outputs/char2Idx.npy"
charembedding_path2 = "../Model/outputs/char2Idx_caseless.npy"
posembedding_path = "../Model/outputs/pos2Idx.npy"


def load_keras_model(modelDIR):
	json_file = open(modelDIR + 'NeuroTPR.json', 'r')
	loaded_model_json = json_file.read()
	json_file.close()
	loaded_model = model_from_json(loaded_model_json, custom_objects={
		'CRF': CRF, 'crf_loss': crf_loss, "crf_viterbi_accuracy": crf_viterbi_accuracy,
		'ElmoEmbeddingLayer': ElmoEmbeddingLayer})
	loaded_model.load_weights(modelDIR + 'NeuroTPR.h5')
	print("Loaded NeuroTPR model from disk")
	return loaded_model


def tag_article(input, test_article, model, iinx, binx):

	result_lines = ""

	for i, sentence in enumerate(input):
		## per sentence as the input
		tokens, chars, chars2, poss, texts = sentence
		tokens = np.asarray([tokens])
		chars = np.asarray([chars])
		chars2 = np.asarray([chars2])
		poss = np.asarray([poss])
		texts = np.asarray([texts])

		if len(texts[0]) < 5:
			print("Too short input")
			break

		pred = model.predict([tokens, chars, chars2, poss, texts], verbose=False)[0] # 

		pred = pred.argmax(axis=-1)  # Predict the classes

		# print(pred)
		tokenIdx = 0
		result_line = ""
		while tokenIdx < len(pred):
			toponym = ""
			if pred[tokenIdx] == binx:  # A new chunk starts
				old_idx = tokenIdx
				toponym = test_article[i][0][tokenIdx]
				tokenIdx += 1

				while tokenIdx < len(pred) and pred[tokenIdx] == iinx:
					toponym += (" " + test_article[i][0][tokenIdx])
					tokenIdx += 1

				if toponym.endswith(","):
					toponym = toponym[:-2]

				if toponym.find("Harvey") == -1 and len(toponym) > 1:
					result_line += toponym + ",,"
					result_line += str(test_article[i][1][old_idx]) + ",," + str(
						test_article[i][1][old_idx] + len(toponym) - 1) + "||"

			elif pred[tokenIdx] == iinx:
				toponym = test_article[i][0][tokenIdx]
				print(toponym)
				tokenIdx += 1

			else:
				tokenIdx += 1

		result_lines += result_line 

	return result_lines


def tag_corpus(model, article_path, word2Idx, idx2Label, char2Idx, char2Idx_caseless, pos2Idx):

	bloc_index = idx2Label["B-location"]
	iloc_index = idx2Label["I-location"]

	sentences = readfile_nolabel(article_path)

	new_sentences = copy.deepcopy(sentences)

	sentences_char = addCharInformatioin(sentences)

	processed_sentences = padding(createMatrices_nolabel_char(sentences_char, word2Idx, char2Idx,
															  char2Idx_caseless, pos2Idx))
	sentences = build_senteceMatrix(new_sentences)

	temp = tag_article(processed_sentences, sentences, model, iloc_index, bloc_index)

	print("The toponym recognition results are:")
	print(temp)
	print("Over")


print("Load the word and linguistic embeddings from npy files")	
model = load_keras_model(model_path)

print("Load the word and linguistic embeddings from npy files")
word2Idx = np.load(wordembedding_path, allow_pickle=True).item()
labelIdx = np.load(labelset_path, allow_pickle=True).item()
char2Idx = np.load(charembedding_path, allow_pickle=True).item()
char2Idx_caseless = np.load(charembedding_path2, allow_pickle=True).item()
pos2Idx = np.load(posembedding_path, allow_pickle=True).item()
inv_labelIdx = {v: k for k, v in labelIdx.items()}

print("processing on the input texts now ...")
tag_corpus(file_path, word2Idx, inv_labelIdx, char2Idx, char2Idx_caseless, pos2Idx)



import numpy as np
import random
from keras.preprocessing.sequence import pad_sequences


def readfile(filename):
    '''
    read file
    return format :
    [ ['EU', 'B-ORG'], ['rejects', 'O'], ['German', 'B-MISC'], ['call', 'O'], ['to', 'O'], ['boycott', 'O'], ['British', 'B-MISC'], ['lamb', 'O'], ['.', 'O'] ]
    '''
    f = open(filename)
    sentences = []
    sentence = []
    for line in f:
        line = line.rstrip()
        if len(line) == 0:
            if len(sentence) > 0:
                sentences.append(sentence)
                sentence = []
            continue
        splits = line.split('\t')
        if len(splits[0]) == 0:
            continue
        sentence.append([splits[0], splits[1], splits[2], splits[3]])


    if len(sentence) > 0:
        sentences.append(sentence)
        sentence = []
    return sentences


def createBatches(data):
    l = []
    for i in data:
        l.append(len(i[0]))
    l = list(set(l))
    random.shuffle(l)
    batches = []
    batch_len = []
    z = 0
    for i in l:
        for batch in data:
            if len(batch[0]) == i:
                batches.append(batch)
                z += 1
        batch_len.append(z)
    return batches, batch_len


def addCharInformatioin(Sentences):
    for i, sentence in enumerate(Sentences):
        for j, data in enumerate(sentence):
            chars = [c for c in data[0]]
            Sentences[i][j] = [data[0], chars, data[1], data[2], data[3]]
    return Sentences


def padding(Sentences):
    maxlen = 52
    for sentence in Sentences:
        char = sentence[1]
        for x in char:
            maxlen = max(maxlen, len(x))
    for i, sentence in enumerate(Sentences):
        Sentences[i][1] = pad_sequences(Sentences[i][1], 52, padding='post')
        Sentences[i][2] = pad_sequences(Sentences[i][2], 52, padding='post')
    return Sentences


def createMatrices_char(sentences, word2Idx, label2Idx, char2Idx, char2Idx_caseless, pos2Idx):
    unknownIdx = word2Idx['UNKNOWN_TOKEN']
    paddingIdx = word2Idx['PADDING_TOKEN']

    dataset = []

    wordCount = 0
    unknownWordCount = 0

    for sentence in sentences:
        wordIndices = []
        charIndices = []
        charIndices2 = []
        labelIndices = []
        posIndices = []
        # nerIndices = []
        wordStrings = ""

        for word, char, label, ner, pos in sentence:
            wordCount += 1
            wordStrings = wordStrings + word + " "

            if word in word2Idx:
                wordIdx = word2Idx[word]
            elif word.lower() in word2Idx:
                wordIdx = word2Idx[word.lower()]
            else:
                wordIdx = unknownIdx
                unknownWordCount += 1
            charIdx1 = []
            charIdx2 = []
            for x in char:
                if x not in char2Idx:
                    x = "UNKNOWN"
                charIdx1.append(char2Idx[x])
                charIdx2.append(char2Idx_caseless[x])
            # Get the label and map to int
            wordIndices.append(wordIdx)
            charIndices.append(charIdx1)
            charIndices2.append(charIdx2)
            labelIndices.append(label2Idx[label])
            posIndices.append(pos2Idx[pos])
            # nerIndices.append(ner2Idx[ner])

        dataset.append([wordIndices, charIndices, charIndices2, labelIndices, posIndices, wordStrings[:-1]])

    return dataset


def iterate_minibatches_char(dataset, batch_len):
    start = 0
    for i in batch_len:
        tokens = []
        char = []
        char2 = []
        labels = []
        words = []
        poss = []
        # ners = []
        data = dataset[start:i]
        start = i
        for dt in data:
            t, ch, ch2, l, p, word = dt
            l = np.expand_dims(l, -1)
            tokens.append(t)
            char.append(ch)
            char2.append(ch2)
            labels.append(l)
            words.append(word)
            poss.append(p)
            # ners.append(n)

        yield np.asarray(labels), np.asarray(tokens), np.asarray(char), np.asarray(char2),\
            np.asarray(poss), np.array(words, dtype=object)[:, np.newaxis]




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="neurotpr", 
    version="0.0.2",
    author="Jimin Wang and Yingjie Hu",
    author_email="yhu42@buffalo.edu",
    description="A Neuro-net ToPonym Recognition model",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/geoai-lab/NeuroTPR",
    packages=setuptools.find_packages(),
    install_requires=[
          'tensorflow',
          'keras',
          'tensorflow-hub',
          'emoji',
          'nltk'
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)